/*
 * Copyright (c) 1999 Greg Haerr <greg@censoft.com>
 * Copyright (c) 1991 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Device-independent top level mouse and cursor routines
 *
 * Reads data from mouse driver and tracks real position on the screen.
 * Intersection detection for cursor with auto removal
 */
#include "device.h"

static COORD	xpos;		/* current x position of mouse */
static COORD	ypos;		/* current y position of mouse */
static COORD	minx;		/* minimum allowed x position */
static COORD	maxx;		/* maximum allowed x position */
static COORD	miny;		/* minimum allowed y position */
static COORD	maxy;		/* maximum allowed y position */
static int	scale;		/* acceleration scale factor */
static int	thresh;		/* acceleration threshhold */
static BUTTON	buttons;	/* current state of buttons */
static BOOL	changed;	/* mouse state has changed */

static COORD 	curminx;	/* minimum x value of cursor */
static COORD 	curminy;	/* minimum y value of cursor */
static COORD 	curmaxx;	/* maximum x value of cursor */
static COORD 	curmaxy;	/* maximum y value of cursor */
static BOOL	curvisible;	/* TRUE if cursor is visible*/
static BOOL 	curneedsrestore;/* cursor needs restoration after drawing*/
static COORD 	cursavx;	/* saved cursor location*/
static COORD 	cursavy;
static COORD	cursavx2;
static COORD	cursavy2;
static PIXELVAL curfg;		/* foreground color of cursor */
static PIXELVAL curbg;		/* background color of cursor */
static PIXELVAL cursavbits[MAX_CURSOR_SIZE * MAX_CURSOR_SIZE];
static IMAGEBITS cursormask[MAX_CURSOR_SIZE];
static IMAGEBITS cursorcolor[MAX_CURSOR_SIZE];

extern MODE gr_mode;

/*
 * Initialize the mouse.
 * This sets its position to (0, 0) with no boundaries and no buttons pressed.
 * Returns < 0 on error, or mouse fd on success
 */
int
GdOpenMouse(void)
{
	int fd;

	if ((fd = mousedev.Open(&mousedev)) < 0)
		return -1;

	/* get default acceleration settings*/
	mousedev.GetDefaultAccel(&scale, &thresh);

	/* init mouse position info*/
	buttons = 0;
	xpos = 0;
	ypos = 0;
	minx = COORD_MIN;
	miny = COORD_MIN;
	maxx = COORD_MAX;
	maxy = COORD_MAX;
	changed = TRUE;

	/* init cursor position and size info*/
	curvisible = FALSE;
	curneedsrestore = FALSE;
	curminx = minx;
	curminy = miny;
	curmaxx = curminx + MAX_CURSOR_SIZE - 1;
	curmaxy = curminy + MAX_CURSOR_SIZE - 1;

	return fd;
}

/*
 * Terminate the use of the mouse.
 */
void
GdCloseMouse(void)
{
	mousedev.Close();
}

void
GdGetButtonInfo(BUTTON *buttons)
{
	*buttons = mousedev.GetButtonInfo();
}

/*
 * Restrict the coordinates of the mouse to the specified coordinates.
 */
void
GdRestrictMouse(COORD newminx, COORD newminy, COORD newmaxx, COORD newmaxy)
{
	minx = newminx;
	miny = newminy;
	maxx = newmaxx;
	maxy = newmaxy;
	GdMoveMouse(xpos, ypos);
}

/*
 * Set the acceleration threshhold and scale factors.
 * Acceleration makes the cursor move further for faster movements.
 * Basically, at mouse speeds above the threshold, the excess distance
 * moved is multiplied by the scale factor.  For example, with a threshhold
 * of 5 and a scale of 3, the following gives examples of the original and
 * modified mouse movements:
 *	input:		0	4	5	6	9	20
 *	output:		0	4	5	8	17	50
 */
void
GdSetAccelMouse(int newthresh, int newscale)
{
	if (newthresh < 0)
		newthresh = 0;
	if (newscale < 0)
		newscale = 0;
	thresh = newthresh;
	scale = newscale;
}

/*
 * Move the mouse to the specified coordinates.
 * The location is limited by the current mouse coordinate restrictions.
 */
void
GdMoveMouse(COORD newx, COORD newy)
{
	if (newx < minx)
		newx = minx;
	if (newx > maxx)
		newx = maxx;
	if (newy < miny)
		newy = miny;
	if (newy > maxy)
		newy = maxy;
	if (newx == xpos && newy == ypos)
		return;

	changed = TRUE;
	xpos = newx;
	ypos = newy;
}

/*
 * Read the current location and button states of the mouse.
 * Returns -1 on read error, 0 if nothing new has happened, and 1 if a
 * new position was read.  The current mouse position and buttons are
 * returned even if the position hasn't changed.  This routine does
 * not block.
 */
int
GdReadMouse(COORD *xp, COORD *yp, BUTTON *bp)
{
	COORD	deltax;		/* change in x position */
	COORD	deltay;		/* change in y position */
	BUTTON	newbuttons;	/* new button state */
	int	sign;		/* sign of change */
	int	status;		/* status of reading mouse */

	*xp = xpos;
	*yp = ypos;
	*bp = buttons;

	if (changed) {
		changed = FALSE;
		return 1;
	}

	status = mousedev.Read(&deltax, &deltay, &newbuttons);
	if (status < 0)
		return -1;

	if (status == 0)
		return 0;

	if (buttons != newbuttons) {
		changed = TRUE;
		buttons = newbuttons;
	}

	sign = 1;
	if (deltax < 0) {
		sign = -1;
		deltax = -deltax;
	}
	if (deltax > thresh)
		deltax = thresh + (deltax - thresh) * scale;
	deltax *= sign;

	sign = 1;
	if (deltay < 0) {
		sign = -1;
		deltay = -deltay;
	}
	if (deltay > thresh)
		deltay = thresh + (deltay - thresh) * scale;
	deltay *= sign;

	GdMoveMouse(xpos + deltax, ypos + deltay);

	if (!changed)
		return 0;

	changed = FALSE;
	*xp = xpos;
	*yp = ypos;
	*bp = buttons;
	return 1;
}

/*
 * Set the cursor position.
 */
void
GdMoveCursor(COORD newx, COORD newy)
{
	COORD shiftx;
	COORD shifty;

	shiftx = newx - curminx;
	shifty = newy - curminy;
	if(shiftx == 0 && shifty == 0)
		return;
	curminx += shiftx;
	curmaxx += shiftx;
	curminy += shifty;
	curmaxy += shifty;

	/* Restore the screen under the mouse pointer*/
	GdHideCursor();

	/* Draw the new pointer*/
	GdShowCursor();
}

/*
 * Set the cursor size and bitmaps.
 */
void
GdSetCursor(COORD width, COORD height, COLORVAL foreground,
	COLORVAL background, IMAGEBITS *fgbitmap, IMAGEBITS *bgbitmap)
{
	BOOL	savevisible;
	int	bytes;

	if (width <= 0 || width > MAX_CURSOR_SIZE ||
	    height <= 0 || height > MAX_CURSOR_SIZE)
		return;

	savevisible = GdHideCursor();
	curfg = GdFindColor(foreground);
	curbg = GdFindColor(background);
	curmaxx = curminx + width - 1;
	curmaxy = curminy + height - 1;
	bytes = IMAGE_WORDS(width) * height * sizeof(IMAGEBITS);
	memcpy(cursorcolor, fgbitmap, bytes);
	memcpy(cursormask, bgbitmap, bytes);
	if (savevisible)
		GdShowCursor();
}


/*
 * Draw the mouse pointer.  Save the screen contents underneath
 * before drawing. Returns previous cursor state.
 */
int
GdShowCursor(void)
{
	COORD 		x;
	COORD 		y;
	PIXELVAL *	saveptr;
	IMAGEBITS *	cursorptr;
	IMAGEBITS *	maskptr;
	IMAGEBITS 	curbit, cbits, mbits;
	PIXELVAL 	oldcolor;
	PIXELVAL 	newcolor;
	MODE 		oldmode;

	if(curvisible)
		return TRUE;
	oldmode = gr_mode;
	gr_mode = MODE_SET;

	saveptr = cursavbits;
	cursavx = curminx;
	cursavy = curminy;
	cursavx2 = curmaxx;
	cursavy2 = curmaxy;
	cursorptr = cursorcolor;
	maskptr = cursormask;

	for (y = curminy; y <= curmaxy; y++) {
		cbits = *cursorptr++;
		mbits = *maskptr++;
		curbit = IMAGE_FIRSTBIT;
		for (x = curminx; x <= curmaxx; x++) {
			if(x >= 0 && x < scrdev.xres &&
			   y >= 0 && y < scrdev.yres) {
				oldcolor = scrdev.ReadPixel(x, y);
				if (curbit & mbits) {
					newcolor = (curbit&cbits)? curbg: curfg;
					if (oldcolor != newcolor)
					       scrdev.DrawPixel(x, y, newcolor);
				}
				*saveptr++ = oldcolor;
			}
			curbit = IMAGE_NEXTBIT(curbit);
		}
	}

	gr_mode = oldmode;
	curvisible = TRUE;
	curneedsrestore = FALSE;
	return FALSE;
}

/*
 * Restore the screen overwritten by the cursor.
 */
int
GdHideCursor(void)
{
	PIXELVAL *	saveptr;
	COORD 		x, y;
	MODE 		oldmode;

	if(!curvisible)
		return FALSE;
	oldmode = gr_mode;
	gr_mode = MODE_SET;

	saveptr = cursavbits;
	for (y = cursavy; y <= cursavy2; y++) {
		for (x = cursavx; x <= cursavx2; x++) {
			if(x >= 0 && x < scrdev.xres && 
			   y >= 0 && y < scrdev.yres) {
				scrdev.DrawPixel(x, y, *saveptr++);
			}
		}
	}
 	gr_mode = oldmode;
	curvisible = FALSE;
	return TRUE;
}

/* Check to see if the mouse pointer is about to be overwritten.
 * If so, then remove the cursor so that the graphics operation
 * works correctly.  If the cursor is removed, then this fact will
 * be remembered and a later call to GdFixCursor will restore it.
 */
void
GdCheckCursor(COORD x1,COORD y1,COORD x2,COORD y2)
{
  COORD temp;

  if(curneedsrestore)
	  return;

  if (x1 > x2) {
	temp = x1;
	x1 = x2;
	x2 = temp;
  }
  if (y1 > y2) {
	temp = y1;
	y1 = y2;
	y2 = temp;
  }
  if ((x1 > curmaxx) || (x2 < curminx) ||
      (y1 > curmaxy) || (y2 < curminy))
	return;

  GdHideCursor();
  curneedsrestore = 1;
}


/* Redisplay the cursor if it was removed because of a graphics operation. */
void
GdFixCursor(void)
{
	if (curneedsrestore)
		GdShowCursor();
}
